//-*- mode: Verilog; verilog-indent-level: 3; indent-tabs-mode: nil; tab-width: 1 -*-

//****************************************************************************************
interface rst_intf;
   logic rst_n;
endinterface : rst_intf

//****************************************************************************************
interface my_intf(input logic clk,
		  input logic rst_n);

   logic 		      sig;

   clocking mon_cb @(posedge clk);
      input 		      sig;
   endclocking : mon_cb

   clocking drv_cb @(posedge clk);
      output 		      sig;
   endclocking : drv_cb
   
   modport mon_mp(clocking mon_cb, input rst_n);
   modport drv_mp(clocking drv_cb, input rst_n);
endinterface : my_intf

//****************************************************************************************
// class: rst_drv_c
class rst_drv_c extends uvm_driver;
   `uvm_component_utils_begin(rst_drv_c)
      `uvm_field_string(intf_name,            UVM_ALL_ON)
      `uvm_field_int(reset_time_ps,           UVM_ALL_ON | UVM_DEC)
   `uvm_component_utils_end

   // var: intf_name
   string 		      intf_name = "rst_i";
   
   // var: reset_time_ps
   // The length of time, in ps, that reset will stay active
   rand int reset_time_ps;

   // Base constraints
   constraint rst_cnstr { reset_time_ps inside {[1:1000000]}; }

   // var: rst_vi
   // Reset virtual interface
   virtual 		      rst_intf rst_vi;

   ////////////////////////////////////////////
   function new(string name="rst_drv", 
		uvm_component parent=null);
      super.new(name, parent);
   endfunction : new

   ////////////////////////////////////////////
   virtual 		      function void build_phase(uvm_phase phase);
      super.build_phase(phase);
      // get the interface
      uvm_resource_db#(virtual rst_intf)::read_by_name("rst_intf", intf_name, rst_vi);
   endfunction : build_phase

   ////////////////////////////////////////////
   virtual 		      task reset_phase(uvm_phase phase);
      phase.raise_objection(this);
      `uvm_info("rst", "Going down.", UVM_NONE)
      rst_vi.rst_n <= 0;
      #(reset_time_ps * 1ps);
      rst_vi.rst_n <= 1;
      `uvm_info("rst", "Going up.", UVM_NONE)
      phase.drop_objection(this);
   endtask : reset_phase
endclass : rst_drv_c

//****************************************************************************************
class item_c extends uvm_sequence_item;
   `uvm_object_utils_begin(item_c)
   `uvm_object_utils_end

   //----------------------------------------------------------------------------------------
   // Group: Fields

   rand int delay;
   constraint delay_cnstr { delay inside {[0:6]}; }
   
   //----------------------------------------------------------------------------------------
   // Group: Methods
   function new(string name="item");
      super.new(name);
   endfunction : new
endclass : item_c

//****************************************************************************************
class mon_c extends uvm_monitor;
   `uvm_component_utils_begin(mon_c)
      `uvm_field_string(intf_name, UVM_DEFAULT)
   `uvm_component_utils_end

   string intf_name;
   virtual my_intf.mon_mp my_vi;

   ////////////////////////////////////////////
   // func: new
   function new(string name="mon",
		uvm_component parent=null);
      super.new(name, parent);
   endfunction : new         
   
   ////////////////////////////////////////////
   // func: build_phase
   virtual function void build_phase(uvm_phase phase);
      super.build_phase(phase);
      // get the interface
      uvm_resource_db#(virtual my_intf.mon_mp)::read_by_name("my_intf", {intf_name, ".mon_mp"}, my_vi);
   endfunction : build_phase
   
   ////////////////////////////////////////////
   virtual task run_phase(uvm_phase phase);
      forever begin
	 @(posedge my_vi.rst_n);
	 
	 fork
            monitor_items();
	 join_none
	 
	 @(negedge my_vi.rst_n);
	 disable fork;
      end
   endtask : run_phase

   ////////////////////////////////////////////
   task monitor_items();
      forever begin
	 @(my_vi.mon_cb.sig);
	 `uvm_info("mon", $sformatf("sig is a %0b.", my_vi.mon_cb.sig), UVM_NONE)
      end
   endtask : monitor_items
endclass : mon_c

//****************************************************************************************
class drv_c extends uvm_driver#(item_c);
   `uvm_component_utils_begin(drv_c)
      `uvm_field_string(intf_name, UVM_DEFAULT)
   `uvm_component_utils_end

   string intf_name;
   event  reset_driver;

   virtual my_intf.drv_mp my_vi;

   ////////////////////////////////////////////
   // func: new
   function new(string name="drv",
		uvm_component parent=null);
      super.new(name, parent);
   endfunction : new         

   ////////////////////////////////////////////
   // func: build_phase
   virtual function void build_phase(uvm_phase phase);
      
      super.build_phase(phase);
      uvm_resource_db#(virtual my_intf.drv_mp)::read_by_name("my_intf", {intf_name, ".drv_mp"}, my_vi);
   endfunction : build_phase

   ////////////////////////////////////////////
   virtual task run_phase(uvm_phase phase);
      forever begin
	 reset();
         @(posedge my_vi.rst_n);
         fork
            driver();
         join_none

         @(reset_driver);
         disable fork;
	 end
   endtask : run_phase

   ////////////////////////////////////////////
   virtual task driver();
      bit value;

      forever begin
	 seq_item_port.get_next_item(req);
	 `uvm_info("drv", $sformatf("Inverting sig after %0d clocks.", req.delay), UVM_INFO)
	 repeat(req.delay)
	    @(my_vi.drv_cb);
	 value = ~value;
	 my_vi.drv_cb.sig <= value;
	 seq_item_port.item_done(req);
      end
   endtask : driver

   ////////////////////////////////////////////
   function void reset();
      my_vi.drv_cb.sig <= 0;
   endfunction : reset
endclass : drv_c

//****************************************************************************************
class seq_c extends uvm_sequence #(item_c);
   `uvm_object_utils(seq_c)

   function new(string name="seq");
      super.new(name);
   endfunction : new

   ////////////////////////////////////////////
   // func: body
   virtual task body();
      item_c item;
      repeat(100) begin
	 `uvm_do(item)
	 get_response(rsp);
      end
   endtask : body
endclass : seq_c

//****************************************************************************************
class sqr_c extends uvm_sequencer#(item_c,item_c);
   `uvm_component_utils(sqr_c)

   function new(string name="sqr",
                uvm_component parent=null);
      super.new(name, parent);
   endfunction : new

   ////////////////////////////////////////////
   // func: main_phase
   virtual task main_phase(uvm_phase phase);
      seq_c seq;

      phase.raise_objection(this);
      seq = seq_c::type_id::create("seq");
      seq.randomize();
      seq.start(this);
      phase.drop_objection(this);
   endtask : main_phase
endclass : sqr_c

//****************************************************************************************
class agent_c extends uvm_agent;
   `uvm_component_utils(agent_c)

   sqr_c sqr;
   drv_c drv;
   mon_c mon;

   //----------------------------------------------------------------------------------------
   // Group: Methods
   function new(string name="agent",
                uvm_component parent=null);
      super.new(name, parent);
   endfunction : new

   ////////////////////////////////////////////
   virtual function void build_phase(uvm_phase phase);
      super.build_phase(phase);
      sqr = sqr_c::type_id::create("sqr", this);
      drv = drv_c::type_id::create("drv", this);
      mon = mon_c::type_id::create("mon", this);
   endfunction : build_phase

   ////////////////////////////////////////////
   virtual function void connect_phase(uvm_phase phase);
      super.connect_phase(phase);
      if(sqr && drv)
	 drv.seq_item_port.connect(sqr.seq_item_export);
   endfunction : connect_phase

   ////////////////////////////////////////////
   virtual task pre_reset_phase(uvm_phase phase);
      if(sqr && drv) begin
         sqr.stop_sequences();
         ->drv.reset_driver;
      end
   endtask : pre_reset_phase
endclass : agent_c

//****************************************************************************************
class base_test_c extends uvm_test;
   `uvm_component_utils(base_test_c)

   agent_c p_agent, b_agent;
   rst_drv_c p_rst_drv, b_rst_drv;

   //----------------------------------------------------------------------------------------
   // Group: Methods
   function new(string name="base_test",
                uvm_component parent=null);
      super.new(name, parent);
   endfunction : new

   ////////////////////////////////////////////
   // func: build_phase
   virtual function void build_phase(uvm_phase phase);
      super.build_phase(phase);

      p_agent = agent_c::type_id::create("p_agent", this);
      uvm_config_db#(string)::set(this, "p_agent.*", "intf_name", "p_domain");

      b_agent = agent_c::type_id::create("b_agent", this);
      uvm_config_db#(string)::set(this, "b_agent.*", "intf_name", "b_domain");

      p_rst_drv = rst_drv_c::type_id::create("p_rst_drv", this);
      uvm_config_db#(string)::set(this, "p_rst_drv", "intf_name", "p_rst_i");
      b_rst_drv = rst_drv_c::type_id::create("b_rst_drv", this);
      uvm_config_db#(string)::set(this, "b_rst_drv", "intf_name", "b_rst_i");

      p_rst_drv.randomize();
      b_rst_drv.randomize();
   endfunction : build_phase

   ////////////////////////////////////////////
   virtual task configure_phase(uvm_phase phase);
      phase.raise_objection(this);
      #(100ns);
      phase.drop_objection(this);
   endtask : configure_phase

endclass : base_test_c

//****************************************************************************************
class idle_reset_test_c extends base_test_c;
   `uvm_component_utils(idle_reset_test_c)

   // field: run_count
   // The number of times the test has run so far
   int run_count;
   
   function new(string name="idle_reset", uvm_component parent=null);
      super.new(name, parent);
   endfunction : new

   virtual function void phase_ready_to_end(uvm_phase phase);
      super.phase_ready_to_end(phase);
      if(phase.get_imp() == uvm_shutdown_phase::get()) begin

	 if(run_count == 0) begin
	    `uvm_info("test", "Hitting Reset!!!", UVM_NONE)
            phase.jump(uvm_pre_reset_phase::get());
            run_count++;
	 end
      end
   endfunction : phase_ready_to_end
endclass : idle_reset_test_c

//****************************************************************************************
class active_reset_test_c extends base_test_c;
   `uvm_component_utils(active_reset_test_c)

   // field: hit_reset
   // Clear this after the reset event to ensure that it only happens once
   bit hit_reset = 1;

   // field: reset_delay_ns
   // The amount of time, in ns, before applying reset during the main phase
   int unsigned reset_delay_ns;

   function new(string name="active_reset", uvm_component parent=null);
      super.new(name, parent);
   endfunction : new

   virtual task main_phase(uvm_phase phase);
      fork
         super.main_phase(phase);
      join_none

      if(hit_reset) begin
         phase.raise_objection(this);
         std::randomize(reset_delay_ns) with { reset_delay_ns inside {[200:400]}; };
         #(reset_delay_ns * 1ns);
         phase.drop_objection(this);
         phase.get_objection().set_report_severity_id_override(UVM_WARNING, "OBJTN_CLEAR", UVM_INFO);
	 `uvm_info("test", "Hitting Reset!!!", UVM_NONE)
         phase.jump(uvm_pre_reset_phase::get());
         hit_reset = 0;
      end
   endtask : main_phase
endclass : active_reset_test_c

//****************************************************************************************
class domain_reset_test_c extends base_test_c;
   `uvm_component_utils(domain_reset_test_c)

   int run_count;
   
   uvm_domain p_domain;

   rand int reset_delay_ns;
   constraint delay_cnstr { reset_delay_ns inside {[100:400]}; }
   
   function new(string name="domain_reset", uvm_component parent=null);
      super.new(name, parent);
   endfunction : new

   ////////////////////////////////////////////
   virtual function void build_phase(uvm_phase phase);
      super.build_phase(phase);

      p_domain = new("p_domain");

      // assign the p_agent, and all its sub-components, to the p_domain
      p_agent.set_domain(p_domain, .hier(1));
      p_rst_drv.set_domain(p_domain);
   endfunction : build_phase

   ////////////////////////////////////////////
   virtual task main_phase(uvm_phase phase);
      fork
         super.main_phase(phase);
      join_none
      
      if(run_count == 0) begin
         phase.raise_objection(this);
         randomize();
         #(reset_delay_ns * 1ns);
         phase.drop_objection(this);
         phase.get_objection().set_report_severity_id_override(UVM_WARNING, "OBJTN_CLEAR", UVM_INFO);
	 `uvm_info("test", "Hitting Reset on p_domain!!!", UVM_NONE)
         p_domain.jump(uvm_pre_reset_phase::get());
         run_count++;
      end
   endtask : main_phase
endclass : domain_reset_test_c

//****************************************************************************************
module tb;
   logic clk;
   
   rst_intf p_rst_i();
   rst_intf b_rst_i();

   my_intf  my_p_i(clk, p_rst_i.rst_n);
   my_intf  my_b_i(clk, b_rst_i.rst_n);
   
   initial begin
      uvm_resource_db#(virtual rst_intf)::set("rst_intf", "p_rst_i", p_rst_i);
      uvm_resource_db#(virtual rst_intf)::set("rst_intf", "b_rst_i", b_rst_i);

      uvm_resource_db#(virtual my_intf.mon_mp)::set("my_intf", "p_domain.mon_mp", my_p_i.mon_mp);
      uvm_resource_db#(virtual my_intf.drv_mp)::set("my_intf", "p_domain.drv_mp", my_p_i.drv_mp);

      uvm_resource_db#(virtual my_intf.mon_mp)::set("my_intf", "b_domain.mon_mp", my_b_i.mon_mp);
      uvm_resource_db#(virtual my_intf.drv_mp)::set("my_intf", "b_domain.drv_mp", my_b_i.drv_mp);
      run_test();
   end

   initial begin
      clk = 0;
      forever begin
	 #(1ns) clk = ~clk;
      end
   end
endmodule : tb
